<?php
// $Id: adserve.inc,v 1.1.2.14 2008/01/08 14:59:03 jeremy Exp $

/**
 * Configuration.
 *
 * There is currently only one configuration option, so it lives here.  If we
 * add additional configuration options, we may at that time move them into an
 * external configuration file.
 */

// Optional, hard code the path to Drupal's root directory.
//define('DRUPAL_ROOT', '/var/www/html');

/**
 * End of configuration.
 */

/**********/

/**
 * The main adserve logic.
 */
function adserve_ad($options = array()) {
  static $displayed_count = 0;

  // If no $options are passed in, assume we're using JavaScript.
  if (!empty($options)) {
    adserve_variable('variable_load', $options);
  }
  $method = adserve_variable('method');

  adserve_include_drupal();
  adserve_debug();

  adserve_variable('error', FALSE);
  $output = NULL;
  if (adserve_variable('cache') != 'none') {
    /**
     * Ad caches are defined through external modules.  Ad caches are composed 
     * of a module 'ad_cache_TYPE.module' and an include file 
     * 'ad_cache_TYPE.inc' that live in the 'cache/TYPE' subdirectory where 
     * 'TYPE' is replaced with the type of cache.  For example, the included 
     * file cache lives in 'cache/file'.
     *
     * The ad_cache_TYPE.inc file must have a function named ad_cache_TYPE()
     * which is used to display ads.  It can optionally include a function
     * titled ad_cache_TYPE_variables used to extract any necessary
     * variables from the global $_GET array (this can also be used to override
     * values that would normally be set from $_GET).  Any functions used
     * by this code without bootstrapping Drupal should also be in this file.
     *
     * The ad_cache_TYPE.module file should define the drupal _help() hook
     * so the module can be enabled.  It should also define the _adcacheapi()
     * hook allowing for configuration and processing.  Any functions used by
     * this code after bootstrapping Drupal should also be in this module.
     *
     * Refer to cache/file/* for an implementation example.
     */
    $function = 'ad_cache_'. adserve_variable('cache');
    $output = adserve_invoke($function);

  }

  // If cache fails above (or if no cache is specified) we don't use a cache.
  if ($output == NULL) {
    if (adserve_variable('debug')) {
      echo "No cache enabled.<br />\n";
    }
  
    adserve_bootstrap();
    
    if (adserve_variable('nids')) {
      $id = adserve_variable('nids');
      $nids = explode(',', $id);
      $type = 'nids';
      $ids = adserve_served_ids($type);
  
      // Build query to filter on just selected nids.
      $filter = adserve_build_filter($nids, 'aid');
      $filter2 = adserve_build_filter($ids, 'aid', 'AND', '!=');

      // Randomly determine which ad to display from the specified nids.
      $sql = "SELECT aid, adtype, redirect, activated, maxviews, maxclicks FROM {ads} WHERE adstatus = 'active' $filter $filter2 ORDER BY RAND() LIMIT ". adserve_variable('quantity');
      $result = db_query($sql);

      if (adserve_variable('debug')) {
        echo "Searching for ad from nid list: $id.<br />\n";
        echo "Query: \"$sql;\"<br />\n";
      }
    }
    else if (adserve_variable('tids')) {
      $id = adserve_variable('tids');
      $tids = explode(',', $id);
      $type = 'tids';
      $ids = adserve_served_ids($type);

      // Build query to filter on just selected tids.
      $filter = adserve_build_filter($tids, 'n.tid');
      $filter2 = adserve_build_filter($ids, 'a.aid', 'AND', '!=');
  
      // Randomly determine which ad to display from the specified tids.
      $sql = "SELECT a.aid, a.adtype, a.redirect, a.activated, a.maxviews, a.maxclicks FROM {ads} a INNER JOIN {term_node} n ON a.aid = n.nid WHERE a.adstatus = 'active' $filter $filter2 ORDER BY RAND() LIMIT ". adserve_variable('quantity');
      $result = db_query($sql);

      if (adserve_variable('debug')) {
        echo "Searching for ad from tid list: $id.<br />\n";
        echo "Query: \"$sql;\"<br />\n";
      }
    }
    else {
      $id = 0;
      $type = 'default';
      $ids = adserve_served_ids($type);
      $filter = adserve_build_filter($ids, 'a.aid', 'AND', '!=');

      // Randomly determine which ad to display from those that do not have
      // any tid assigned to them.
      $sql = "SELECT a.aid, a.adtype, a.redirect, a.activated, a.maxviews, a.maxclicks FROM {ads} a LEFT JOIN {term_node} n ON a.aid = n.nid WHERE a.adstatus = 'active' AND n.tid IS NULL $filter ORDER BY RAND() LIMIT ". adserve_variable('quantity');
      $result = db_query($sql);

      if (adserve_variable('debug')) {
        echo "Searching for ads with no tids.<br />\n";
        echo "Query: \"$sql;\"<br />\n";
      }
    }
      
    $output = '';
    $ads = 0;
    // Include appropriate module for displaying selected ad.
    while ($ad = db_fetch_object($result)) {
      $ids[$ad->aid] = $ad->aid;
      $ads++;
      $module = db_result(db_query("SELECT filename FROM {system} WHERE name = '%s'", "ad_$ad->adtype"));
      if (adserve_variable('debug')) {
        echo "ad: <pre>";
        print_r($ad);
        echo "</pre>";
        echo "Loading module '$module'.<br />\n";
      }
      include_once "$module";
      
      if ($output) {
        // Add a div between ads that themers can use to arrange ads when 
        // displaying more than one at a time.
        $displayed_count++;
        $output .= "<div class=\"advertisement-space\" id=\"$id-$displayed_count\"></div>";
      }
      $output .= module_invoke("ad_$ad->adtype", 'display_ad', $ad);

      // Update the ad's view counter.
      if ($method == 'raw') {
        $output .= ad_display_image($ad);
      }
      else {
        adserve_increment($ad);
      }
    }
    adserve_variable("$type-ids", $ids);
    if (empty($ads)) {
      adserve_variable('error', TRUE);
      $output = 'No active ads were found in the '. (empty($nids) ? 'tids' : 'nids') ." '$id'.";
      adserve_increment(NULL, 'count');
    }
    if (adserve_variable('debug')) {
      echo "Ads displayed: $ads<br />";
    }
  }

  $output = preg_replace('&/@HOSTID___&', '/'. adserve_variable('hostid'), $output);
  if (adserve_variable('error')) {
    $output = "<!-- $output -->";
  }

  switch ($method) {
    case 'iframe':
      if (!adserve_variable('debug')) {
        // Tell the web browser not to cache this frame so the ad refreshes 
        // each time the page is viewed.

        // Expires in the past:
        header('Expires: Mon, 26 Jul 1997 05:00:00 GMT');
        // Last load:
        header('Last-Modified: '. gmdate('D, d M Y H:i:s') .' GMT');
        // HTTP 1.1:
        header('Cache-Control: no-store, no-cache, must-revalidate');
        header('Cache-Control: post-check=0, pre-check=0', false);
        // HTTP 1.0:
        header('Pragma: no-cache');
      }
      print "$output";
      exit(0);
    case 'javascript':
    default:
      $output = str_replace(array("\r", "\n", "<", ">", "&"),
                            array('\r', '\n', '\x3c', '\x3e', '\x26'),
                            addslashes($output));
      if (!adserve_variable('debug')) {
        // Tell the web browser not to cache this script so the ad refreshes 
        // each time the page is viewed.
        // Expires in the past:
        header('Expires: Mon, 26 Jul 1997 05:00:00 GMT');
        // Last load:
        header('Last-Modified: '. gmdate('D, d M Y H:i:s') .' GMT');
        // HTTP 1.1:
        header('Cache-Control: no-store, no-cache, must-revalidate');
        header('Cache-Control: post-check=0, pre-check=0', false);
        // HTTP 1.0:
        header('Pragma: no-cache');
        // Output is a JavaScript:
        header('Content-Type: application/x-javascript; charset=utf-8');
      }
      print "document.write('$output');";
      exit(0);
    case 'raw':
      chdir(adserve_variable('ad_dir'));
      return $output;
  }
}

/**
 * Retrieve variables from $_GET array or from passed in $value array.
 */
function adserve_variable($variable, $value = NULL) {
  static $variables = NULL;

  // Update the value, if set.
  if (isset($value)) {
    $variables->$variable = $value;
  }

  if (!isset($variables->loaded) || $variable == 'variable_load') {
    if ($variable == 'variable_load') {
      $values['debug'] = $value['debug'];
      $values['c'] = $value['cache'];
      $values['n'] = $value['nids'];
      $values['t'] = $value['tids'];
      $values['k'] = $value['hostid'];
      $values['q'] = $value['quantity'];
      $values['m'] = $value['method'];
      unset($value);
    }
    else {
      $values = $_GET;
    }

    $variables->ad_dir = getcwd();
    // 'debug' is an integer.
    $variables->debug = isset($values['debug']) ? (int)$values['debug'] : 0;
    // Cache types are comprised of only letters.
    $variables->cache = isset($values['c']) ? preg_replace('/[^a-zA-Z]/', '', $values['c']) : 'none';
    // Nids is an integer or a ",".
    $variables->nids = isset($values['n']) ? preg_replace('/[^0-9,]/', '', $values['n']) : '';
    // Tids is an integer or a ",".
    $variables->tids = isset($values['t']) ? preg_replace('/[^0-9,]/', '', $values['t']) : '';
    // Hostid is an md5() which is comprised of numbers and letters a-f.
    $variables->hostid = isset($values['k']) ? preg_replace('/[^0-9a-f]/', '', $values['k']) : '';
    // Quantity is an integer.
    $variables->quantity = isset($values['q']) ? (int)$values['q'] : 0;
    // Ad ID is an integer.
    $variables->aid = isset($values['a']) ? (int)$values['a'] : 0;
    // Method is compriese of only letters.
    $variables->method = isset($values['m']) ? preg_replace('/[^a-zA-Z]/', '', $values['m']) : 'javascript';

    // Retrieve variables defined by cache plugin, if enabled.
    if ($variables->cache != 'none') {
      $include = $variables->ad_dir ."/cache/$variables->cache/ad_cache_$variables->cache.inc";
      if (file_exists($include)) {
        if ($variables->debug) {
          echo "Attempting to include cache include file '$include'.<br />\n";
        }
        require_once ($include);
      }
      else if ($variables->debug) {
        echo "Failed to find cache include file '$include'.<br />\n";
      }
      $function = "ad_cache_$variables->cache". '_variables';
      if (function_exists($function)) {
        $external_variables = $function();
        foreach ($external_variables as $key => $val) {
          $variables->$key = $val;
        }
      }
    }

    // Set defaults.
    $variables->quantity = $variables->quantity ? $variables->quantity : 1;

    if ($variables->debug) {
      foreach ($variables as $variable => $val) {
        echo "$variable: '$val'<br />\n";
      }
      if ($variables->debug == 1) exit;
    }
    $variables->loaded = TRUE;

    // Override the value, if set during initialization.
    if (isset($value)) {
      $variables->$variable = $value;
    }
  }

  if ($variable == 'debug_dump') {
    echo "Dumping \$variables:<br />\n";
    echo '<pre>';
    foreach ($variables as $variable => $value) {
      echo "  $variable($value)<br />\n";
    }
    echo '</pre>';
  }

  if (isset($variables->$variable)) {
    return $variables->$variable;
  } 
  else {
    return NULL;
  }
}

/**
 * Invoke a function in the specified file.
 */
function adserve_invoke($function, $arg1 = NULL) {
  $output = '';
  if (function_exists("$function")) {
    $output = $function($arg1);
  }
  else if (adserve_variable('debug')) {
    echo "Function '$function' does not exist.<br />\n";
  }
  return $output;
}

function _debug_echo($text) {
  if (adserve_variable('debug')) {
    echo "$text<br />";
  }
}

/**
 * Remove one or more ids from array.
 * TODO: Optimize.  Perhaps something like array_flip, unset, array_flip.
 * @param $ids An array of ID's, ie array(5, 8, 9, 11).
 * @param $remove An ID or an array of ID's to be removed from $ids.
 */
function adserve_select_reindex($ids, $remove) {
  $new = array();
  // Walk through array of IDs and decide what to keep.
  foreach ($ids as $id) {
    // If $remove is an array, walk through array to decide fate of ID.
    if (is_array($remove)) {
      $keep = TRUE;
      foreach ($remove as $rem) {
        // Loop until we find one that matches or reach end of array.
        if ($id == $rem) {
          $keep = FALSE;
          break;
        }
      }
      if ($keep) {
        $new[] = $id;
      }
    }
    else {
      if ($id != $remove) {
        $new[] = $id;
      }
    }
  }
  return $new;
}

/**
 *
 */
function adserve_invoke_weight($ads, $quantity = 1, $invalid = array()) {
  $parent = adserve_variable('ad_dir') .'/weight';
  if (is_dir($parent) && $handle = opendir($parent)) {
    while ($dir = readdir($handle)) {
      if (is_dir("$parent/$dir") && !in_array($dir, array('.', '..', 'CVS'))) {
        $include = "$parent/$dir/ad_weight_$dir.inc";
        if (file_exists($include)) {
          require_once($include);
          $function = "ad_weight_{$dir}_select_ad";
          if (function_exists($function)) {
            $return = $function($ads, $quantity, $invalid);
            // First come, first serve.  We found an ad_weight function that
            // returned something, so we'll take it.
            if ($return) {
              return $return;
            }
          }
        }
      }
    }
  }
}

/**
 * Simple default function to randomly select an ad.  Provides a hook to allow 
 * the definition of external methods.
 * @param An array of valid ad IDs, ie array(5, 8, 9, 11).
 * @param Optional, how many unique ads to select.
 * @param Optional, an array of invalid IDs.
 */
function adserve_select_ad($ads, $quantity = 1, $invalid = array()) {
  adserve_invoke_weight($ads, $quantity, $invalid);

  $ids = array();
  $total = sizeof($ads);
  _debug_echo("Selecting $quantity ad(s) from $total total ad(s).");
  if (is_array($ads)) {
    $ads = adserve_select_reindex($ads, $invalid);
    $total = sizeof($ads);
    for ($i = 0; $i < $quantity; $i++) {
      _debug_echo('Randomly selecting ad: '. ($i + 1) ." of $quantity.");
      // Randomly select a unique banner to display.  We subtract 1 as arrays
      // start at 0.
      $id = $total > 1 ? $ads[mt_rand(0, $total - 1)] : $ads[0];
      _debug_echo("Randomly selected ID: $id.");
      $ids[] = $id;
      $invalid[] = $id;
      $ads = adserve_select_reindex($ads, $id);
      $total = sizeof($ads);
      // We're out of ads to display.
      if ($total <= 0) {
        break;
      }
    }
  }
  return $ids;
}

/**
 * Include Drupal's bootstrap.inc.
 */
function adserve_include_drupal() {
  // Optionally set DRUPAL_ROOT at the top of this file.
  if (DRUPAL_ROOT && is_dir(DRUPAL_ROOT) &&
      file_exists(DRUPAL_ROOT .'/includes/bootstrap.inc')) {
    chdir(DRUPAL_ROOT);
  }
  else if (DRUPAL_ROOT && DRUPAL_ROOT != 'DRUPAL_ROOT' &&
           adserve_variable('debug')) {
    echo 'Invalid DRUPAL_ROOT ('. DRUPAL_ROOT .') defined in adserve.inc.';
  }
  else {
    // Move process to top level Drupal directory so we can perform a bootstrap.
    for ($i = 1; $i <= 20; $i++) {
      if (file_exists('./includes/bootstrap.inc')) {
        break;
      }
      chdir('..');
    }
  }
  adserve_variable('root_dir', getcwd());
  require_once adserve_variable('root_dir') .'/includes/bootstrap.inc';
}

/**
 * Include the necessary files and call the Drupal bootstrap.
 */
function adserve_bootstrap($bootstrap = DRUPAL_BOOTSTRAP_FULL) {
  static $bootstrapped = FALSE;

  if (!$bootstrapped) {
    if (adserve_variable('debug')) {
      echo "Drupal bootstrap '". $bootstrap ."'.<br />\n";
    }

    drupal_bootstrap($bootstrap);
    $bootstrapped = TRUE;
  }
}

/**
 * Increment an ads view counters directly in the database.
 */
function adserve_increment($ad, $action = 'view') {
  adserve_bootstrap();
  // Update view statistics.
  db_query("UPDATE {ad_statistics} SET count = count + 1 WHERE aid = %d AND action = '%s' AND date = %d AND hostid = '%s'", $ad->aid, $action, date('YmdH'), adserve_variable('hostid'));
  // If column doesn't already exist, we need to add it.
  if (!db_affected_rows()) {
    db_query("INSERT INTO {ad_statistics} (aid, date, action, hostid, count) VALUES(%d, %d, '%s', '%s', 1)", $ad->aid, date('YmdH'), $action, adserve_variable('hostid'));
    // If another process already added this row our INSERT will fail, if 
    // so we still need to increment it so we don't loose a view.
    if (!db_affected_rows()) {
      db_query("UPDATE {ad_statistics} SET count = count + 1 WHERE aid = %d AND action = '%s' AND date = %d AND hostid = '%s'", $ad->aid, $action, date('YmdH'), adserve_variable('hostid'));
    }
  }
  
  if ($action == 'view') {
    // See if we need to perform additional queries.
    if ($ad->maxviews) {
      $views = (int)db_result(db_query("SELECT SUM(count) FROM {ad_statistics} WHERE aid = %d AND action = 'view' AND date >= %d", $ad->aid, date('YmdH', $ad->activated)));
      if ($views >= $ad->maxviews) {
        db_query("UPDATE {ads} SET adstatus = 'expired', autoexpire = 0, autoexpired = %d, expired = %d WHERE aid = %d", time(), time(), $ad->aid);
        ad_statistics_increment($ad->aid, 'autoexpired');
        ad_statistics_increment($ad->aid, 'expired');
      }
    }
  }
  // TODO: Do we need to do this here?  Can it happen when a new click is
  // registered?
  if ($ad->maxclicks) {
    $clicks = (int)db_result(db_query("SELECT SUM(count) FROM {ad_statistics} WHERE aid = %d AND action = 'click' AND date >= %d", $ad->aid, date('YmdH', $ad->activated)));
    if ($clicks >= $ad->maxclicks) {
      db_query("UPDATE {ads} SET adstatus = 'expired', autoexpire = 0, autoexpired = %d, expired = %d WHERE aid = %d", time(), time(), $ad->aid);
      ad_statistics_increment($ad->aid, 'autoexpired');
      ad_statistics_increment($ad->aid, 'expired');
    }
  }
}

/**
 * Display additional debug information.
 */
function adserve_debug() {
  if (adserve_variable('debug')) {
    echo "Root drupal directory detected as '". adserve_variable('root_dir') ."'.<br />\n<br />\n";

    $ad_dir = adserve_variable('ad_dir');
    $files = array("$ad_dir/serve.php", "$ad_dir/ad.module");
    if (adserve_variable('debug') > 2) {
      $files = array_merge($files, array("$ad_dir/ad.install", "$ad_dir/ad_views.inc"));
    }
    if (adserve_variable('debug') > 3) {
      $files = array_merge($files, array("$ad_dir/image/ad_image.module", "$ad_dir/image/ad_image.install", "$ad_dir/image/ad_image_views.inc", "$ad_dir/text/ad_text.module", "$ad_dir/text/ad_text.install", "$ad_dir/text/ad_text_views.inc", "$ad_dir/embed/ad_embed.module", "$ad_dir/report/ad_report.module", "$ad_dir/notify/ad_notify.module", "$ad_dir/notify/ad_notify.install"));
    }
    foreach ($files as $file) {
      if (!file_exists($file)) {
        echo "Error: '$file' does not exist!<br />\n";
      }
      else if (!is_readable($file)) {
        echo "Error: '$file' is not readable!<br />\n";
      }
      else {
        $fd = fopen($file, 'r');
        while (!feof($fd)) {
          $line = fgets($fd);
          if (substr($line, 0, 5) == "<?php") {
            continue;
          }
          else {
            echo "$file: $line<br />";
            break;
          }
        }
      }
    }
    echo "<br />\n";
  }
}

/**
 * Get array of all $ids served in current group.
 */
function adserve_served_ids($type = 'default') {
  $ids = adserve_variable("$type-ids");

  if (adserve_variable('debug')) {
    echo 'Ads previously served: '. sizeof($ids) .'<br />';
  }

  if (is_array($ids)) {
    return $ids;
  }
  else {
    return array();
  }
}

/**
 * Build a query filter.
 */
function adserve_build_filter($ids, $column, $andor = 'OR', $comparison = '=') {
  $filter = NULL;

  foreach ($ids as $id) {
    if ($filter) {
      $filter .= " $andor $column $comparison $id";
    }
    else {
      $filter = " AND ($column $comparison $id";
    }
  }
  if ($filter) {
    $filter .= ')';
  }

  return $filter;
}
